﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using YamlDotNet.Serialization;

namespace BendPattern
{
    class Program
    {

        static bool bottom_or_left; // for optimization. If true, next line will be from bottom to top or from left to right. Otherwise, in opposite direction. Inverted after every line.
        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Please provide YAML file with settings");
                return;
            }

            MillSettings settings;

            string fileName = args[0];

            IDeserializer deserializer;

            deserializer = new DeserializerBuilder().Build();
            try
            {
                using (Stream stream = File.OpenRead(fileName))
                {
                    using (TextReader reader = new StreamReader(stream))
                    {
                        settings = deserializer.Deserialize<MillSettings>(reader);
                    }
                }
            }
            catch(Exception ex)
            {
                Console.WriteLine("Error loading settings: " + ex.ToString());
                return;
            }
            if (settings == null)
            {
                Console.WriteLine("Error loading settings");
                return;
            }
            //FileInfo fileInfo = new FileInfo(fileName);
            string target = Path.Combine(Path.GetDirectoryName(fileName),  Path.GetFileNameWithoutExtension(fileName) + ".gcode");
            StringBuilder sb = new StringBuilder();
            bottom_or_left = true;
            if (settings.IsHorizontal)
            {
                int i = 0;
                for (double d=settings.BottomLeftY; d<=settings.TopRightY; d+=settings.StripWidth,i++)
                {
                    double PatternOffset = (i % 2 == 0) ? 0 : - (settings.StrokeLength + settings.GapLength) / 2;
                    MillOneLine(settings, i, sb, PatternOffset);
                }
            }
            else
            {
                int i = 0;
                for (double d = settings.BottomLeftX; d <= settings.TopRightX; d += settings.StripWidth, i++)
                {
                    double PatternOffset = (i % 2 == 0) ? 0 : - (settings.StrokeLength + settings.GapLength) / 2;
                    MillOneLine(settings, i, sb, PatternOffset);
                }
            }
            File.WriteAllText(target, sb.ToString());

        }

        static void MillOneLine(MillSettings settings, int LineNumber, StringBuilder sb, double PatternOffset)
        {
            double Z = 0;
            List<(double x1, double y1, double x2, double y2)> strokes = new List<(double x1, double y1, double x2, double y2)>();
            for (int i = 0; i < settings.ZPasses; i++)
            {
                // first, prepare a list of strokes
                if (settings.IsHorizontal)
                {
                    double x = settings.BottomLeftX + PatternOffset;
                    double y = settings.StripWidth * (double)LineNumber + settings.BottomLeftY;

                    while (x < settings.TopRightX)
                    {
                        double x1 = x;
                        double y1 = y;
                        double x2 = x1 + settings.StrokeLength;
                        double y2 = y;
                        // normalize
                        if (x1 < settings.BottomLeftX)
                        {
                            x1 = settings.BottomLeftX;
                        }
                        if (x2 > settings.TopRightX)
                        {
                            x2 = settings.TopRightX;
                        }

                        // mill
                        if (x2 > x1)
                        {
                            strokes.Add((x1, y1, x2, y2));
                        }
                        // gap
                        x = x2 + settings.GapLength;
                    }
                }
                else
                {
                    double x = settings.StripWidth * (double)LineNumber + settings.BottomLeftX;
                    double y = settings.BottomLeftY + PatternOffset;

                    while (y < settings.TopRightY)
                    {
                        double x1 = x;
                        double y1 = y;
                        double x2 = x;
                        double y2 = y1 + settings.StrokeLength;
                        // normalize
                        if (y1 < settings.BottomLeftY)
                        {
                            y1 = settings.BottomLeftY;
                        }
                        if (y2 > settings.TopRightY)
                        {
                            y2 = settings.TopRightY;
                        }
                        // mill
                        if (y2 > y1)
                        {
                            strokes.Add((x1, y1, x2, y2));
                        }
                        // gap
                        y = y2 + settings.GapLength;
                    }

                }

                // now, mill strokes one by one
                if (bottom_or_left)
                {
                    for(int s=0;s<strokes.Count;s++)
                    {
                        Move(settings, strokes[s].x1, strokes[s].y1, Z, sb);
                        MoveWithMill(settings, strokes[s].x2, strokes[s].y2, Z, sb);
                    }
                }
                else
                {
                    for (int s = strokes.Count - 1; s >=0; s--)
                    {
                        Move(settings, strokes[s].x2, strokes[s].y2, Z, sb);
                        MoveWithMill(settings, strokes[s].x1, strokes[s].y1, Z, sb);
                    }
                }


                Z -= settings.ZStep;
                bottom_or_left = !bottom_or_left;
            }

        }
        static void MoveWithMill(MillSettings settings, double X, double Y, double Z, StringBuilder sb)
        {
            sb.Append($"M3 S{settings.LaserPower}\n");
            sb.Append($"G1 X{X} Y{Y} Z{Z} F{settings.FeedRateMill}\n");
            sb.Append($"M3 S0\n");
        }
        static void Move(MillSettings settings, double X, double Y, double Z, StringBuilder sb)
        {
            sb.Append($"G0 X{X} Y{Y} Z{Z} F{settings.FeedRateMove}\n");
            sb.Append($"\n");
        }
    }
}
